{-# LANGUAGE RankNTypes                       #-}
{-# LANGUAGE GADTs                            #-}
{-# LANGUAGE KindSignatures                   #-}
{-# LANGUAGE ScopedTypeVariables              #-}

module Library.SYB where

import Control.Applicative
import Data.Typeable (Typeable, cast, gcast)

--------------------------------------------------------------------------------
-- SYB
--------------------------------------------------------------------------------

data Spine :: * -> * where
  Con     :: a -> Spine a
  (:<>:)  :: (Data a) => Spine (a -> b) -> a -> Spine b

instance Functor Spine where
  fmap f (Con x)    = Con (f x)
  fmap f (c :<>: x) = fmap (f .) c :<>: x

instance Applicative Spine where
  pure = Con
  Con f      <*> x          = fmap f x
  (c :<>: x) <*> Con y      = fmap (\ f x -> f x y) c :<>: x
  (c :<>: x) <*> (d :<>: y) = (fmap (\ f d y -> f (d y)) (c :<>: x) <*> d) :<>: y

spfold :: (forall a b. Data a => r (a -> b) -> a -> r b)
       -> (forall a. a -> r a)
       -> Spine a
       -> r a
spfold f z (Con c)    = z c
spfold f z (c :<>: x) = spfold f z c `f` x

class Typeable a => Data a where
  spine   :: a -> Spine a
  spine c = Con c
  gfoldl  :: (forall d b. Data d => c (d -> b) -> d -> c b)
          -> (forall g. g -> c g)
          -> a
          -> c a
  gfoldl f z x = spfold f z (spine x)
{-
  gunfold :: (forall b r. Data b => c (b -> r) -> c r)
          -> (forall r. r -> c r)
          -> Constr
          -> c a

  toConstr   :: a -> Constr

  dataTypeOf  :: a -> DataType
-}

type GenericT   = forall a. Data a => a -> a
type GenericQ r = forall a. Data a => a -> r

gmapT :: GenericT -> GenericT
gmapT f x = go (spine x)
  where
    go :: Spine a -> a
    go (Con c)    = c
    go (c :<>: x) = go c (f x)

gmapQr :: forall r' r. (r' -> r -> r) -> r
      -> GenericQ r' -> GenericQ r
gmapQr o r0 f x0 = go (spine x0) r0
  where
    go :: forall a. Spine a -> r -> r
    go (Con c)    y = y
    go (c :<>: x) y = go c (f x `o` y)

gmapQ :: GenericQ r -> GenericQ [r]
gmapQ f = gmapQr (:) [] f

gsize :: GenericQ Int
gsize t = 1 + sum (gmapQ gsize t)

glength :: GenericQ Int
glength = length . gmapQ (const ())

gdepth :: GenericQ Int
gdepth = (+) 1 . foldr max 0 . gmapQ gdepth

fromSpine :: Spine a -> a
fromSpine (Con x)    = x
fromSpine (c :<>: x) = fromSpine c x

--------------------------------------------------------------------------------
-- Stuff involving casts
--------------------------------------------------------------------------------

mkT :: (Typeable a, Typeable b) => (b -> b) -> a -> a
mkT = extT id

mkQ :: (Typeable a, Typeable b) => r -> (b -> r) -> a -> r
(r `mkQ` br) a = case cast a of
                        Just b  -> br b
                        Nothing -> r

-- | Flexible type extension
ext0 :: (Typeable a, Typeable b) => c a -> c b -> c a
ext0 def ext = maybe def id (gcast ext)

-- | Extend a generic transformation by a type-specific case
extT :: (Typeable a, Typeable b) => (a -> a) -> (b -> b) -> a -> a
extT def ext = unT ((T def) `ext0` (T ext))

-- | Extend a generic query by a type-specific case
extQ :: (Typeable a, Typeable b) => (a -> q) -> (b -> q) -> a -> q
extQ f g a = maybe (f a) g (cast a)

-- | The type constructor for transformations
newtype T x = T { unT :: x -> x }

everywhere :: GenericT -> GenericT
everywhere f = f . gmapT (everywhere f)

everything :: (r -> r -> r) -> GenericQ r -> GenericQ r
everything k f x = foldl k (f x) (gmapQ (everything k f) x)
